/*
 * MCP794xxRTC.h - library for MCP794xx RTC/SRAM family of devices
  
  Copyright (c) Manuel Iglesias 2012
  This library is intended to be use with Arduino Time.h library functions and with
  Wire.h library (need to modify BUFFER_LENGTH to allow more than 64 bytes of reception and transmision buffer)
  bufer of the Inter Intergrated Library also need to allow at least 65 bytes to send.

  The library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  
  14 March 2012 - Initial release
 */

#if ARDUINO >= 100
#include <Arduino.h> 
#else
#include <WProgram.h> 
#endif
 
#include <Wire.h>
#include "MCP794xxRTC.h"

#define MCP794xx_RTC_SRAM_CID 0x6F					// RTC and SRAM control ID
#define MCP794xx_UID_EEPROM_CID 0x57				// Unique ID Control ID


MCP794xxRTC::MCP794xxRTC(int part)
{
  Wire.begin();
  _part = part;
  set_address(0);
  _ee_address = 0;							// begin of memory.
}
  
// PUBLIC FUNCTIONS

// RTCC FUNCTIONS

time_t MCP794xxRTC::get()   // Aquire data from buffer and convert to time_t
{
  Serial.println("Time will be synced");
  tmElements_t tm;
  read(tm);
  return(makeTime(tm));
}

void  MCP794xxRTC::set(time_t t)
{
  tmElements_t tm;
  breakTime(t, tm);
  tm.Second = dec2bcd(tm.Second);
  tm.Second &= 0x7f;  // stop the clock 
  write(tm); 
  tm.Second |= 0x80;  // start the clock
  write(tm);

  
}

// Aquire data from the RTC chip in BCD format
void MCP794xxRTC::read( tmElements_t &tm)
{
  Wire.beginTransmission(MCP794xx_RTC_SRAM_CID);
#if ARDUINO >= 100  
  Wire.write((uint8_t)0x00); 
#else
  Wire.send(0x00);
#endif  
  Wire.endTransmission();

  // request the 7 data fields   (secs, min, hr, dow, date, mth, yr)
  Wire.requestFrom(MCP794xx_RTC_SRAM_CID, tmNbrFields);
#if ARDUINO >= 100    
  tm.Second = bcd2dec(Wire.read() & 0x7f);   
  tm.Minute = bcd2dec(Wire.read() );
  tm.Hour =   bcd2dec(Wire.read() & 0x3f);  // mask assumes 24hr clock
  tm.Wday = bcd2dec(Wire.read() & 0x07);	
  tm.Day = bcd2dec(Wire.read() );	
  tm.Month = bcd2dec(Wire.read() & 0x1f);
  tm.Year = y2kYearToTm((bcd2dec(Wire.read())));
#else
  tm.Second = bcd2dec(Wire.receive() & 0x7f);   
  tm.Minute = bcd2dec(Wire.receive() );
  tm.Hour =   bcd2dec(Wire.receive() & 0x3f);  // mask assumes 24hr clock
  tm.Wday = bcd2dec(Wire.receive() & 0x07);
  tm.Day = bcd2dec(Wire.receive() );
  tm.Month = bcd2dec(Wire.receive() & 0x1f);
  tm.Year = y2kYearToTm((bcd2dec(Wire.receive())));
#endif  
}

void MCP794xxRTC::write(tmElements_t &tm)
{
  Wire.beginTransmission(MCP794xx_RTC_SRAM_CID);
#if ARDUINO >= 100  
  Wire.write((uint8_t)0x00); // reset register pointer  
  Wire.write(tm.Second) ;   
  Wire.write(dec2bcd(tm.Minute));
  Wire.write(dec2bcd(tm.Hour));      // sets 24 hour format
  Wire.write(dec2bcd(tm.Wday));   
  Wire.write(dec2bcd(tm.Day));
  Wire.write(dec2bcd(tm.Month));
  Wire.write(dec2bcd(tmYearToY2k(tm.Year))); 
#else  
  Wire.send(0x00); // reset register pointer  
  Wire.send(tm.Second) ;   
  Wire.send(dec2bcd(tm.Minute));
  Wire.send(dec2bcd(tm.Hour));      // sets 24 hour format
  Wire.send(dec2bcd(tm.Wday) | 0x08);   // clear the timestamp bit and set the BACKUP ENABLE BIT!
  Wire.send(dec2bcd(tm.Day));
  Wire.send(dec2bcd(tm.Month));
  Wire.send(dec2bcd(tmYearToY2k(tm.Year)));   
#endif
  Wire.endTransmission();  
}

//SRAM FUNCTIONS
void MCP794xxRTC::read_byte(uint8_t address, uint8_t &byte){
  _address = ((address + 0x20) > 0x5F) ? 0x20 : (address + 0x20);		// SRAM begins at 0x20  
//  _address = address;
  Wire.beginTransmission(MCP794xx_RTC_SRAM_CID);
#if ARDUINO >= 100  
  Wire.write((uint8_t)_address); 
#else
  Wire.send(_address);
#endif  
  Wire.endTransmission();	
  
  // request 1 byte
  Wire.requestFrom(MCP794xx_RTC_SRAM_CID, 1);
  
#if ARDUINO >= 100    
  byte = Wire.read();   
#else
  byte = Wire.receive();   
#endif  
}


void MCP794xxRTC::read_sram(uint8_t* b){
  _address = 0x20;												// this function will leave the internal pointer at this address.
  Wire.beginTransmission(MCP794xx_RTC_SRAM_CID);
#if ARDUINO >= 100  
  Wire.write((uint8_t)_address); 								// SRAM begins at 0x20
#else
  Wire.send(_address);
#endif  
  Wire.endTransmission();	
  
  // request bytes
  Wire.requestFrom(MCP794xx_RTC_SRAM_CID, 64);					// Read the SRAM
  uint8_t i=0;
  while(Wire.available()){
#if ARDUINO >= 100
	b[i] = Wire.read();
#else
	b[i] = Wire.receive();
#endif
	i++;
  }
}

// write a len number of bytes in sequence to the SRAM
int MCP794xxRTC::write_sram(uint8_t address, uint8_t* byte, uint8_t len){
   _address = ((address + 0x20) > 0x5F) ? 0x20 : (address + 0x20);
   Serial.println("write_sram address is: ");
   Serial.println(_address, HEX);
   Wire.beginTransmission(MCP794xx_RTC_SRAM_CID);
 #if ARDUINO >= 100  
   Wire.write((uint8_t)_address); 
   Wire.write((uint8_t)byte, len); 	  
 #else
   Wire.send(_address);
   Wire.send(byte, len);
 #endif  
   _address = ((_address + len) > 0x5F) ? ((_address + len) - 0x40) : (_address + len);		// this will allow to write 64 bytes from an address different that 0
   Wire.endTransmission();
  
   return(1);
}

// write one byte at current address
void MCP794xxRTC::write_byte(uint8_t &byte){
   Wire.beginTransmission(MCP794xx_RTC_SRAM_CID);
 #if ARDUINO >= 100  
   Wire.write((uint8_t)_address); 
   Wire.write((uint8_t)byte); 	  
 #else
   Wire.send(_address);
   Wire.send(byte);
 #endif  
   _address = ((_address + 1) > 0x5F) ? 0x20 : _address + 1;		// this will allow to write 64 bytes from an address different that 0
   Wire.endTransmission();
}

void MCP794xxRTC::write_byte(int byte){
	uint8_t byte2 = (uint8_t)byte;
	write_byte(byte2);
}

void MCP794xxRTC::set_address(uint8_t address){
	_address = ((address + 0x20) > 0x5F) ? 0x20 : (address + 0x20);	
}

void MCP794xxRTC::get_unique_id(uint8_t* uid){
	Wire.beginTransmission(MCP794xx_UID_EEPROM_CID);
#if ARDUINO >= 100  
	Wire.write((uint8_t)0xF0); 								// SRAM begins at 0x20
#else
	Wire.send(0xF0);
#endif  
	Wire.endTransmission();
  // request bytes
  Wire.requestFrom(MCP794xx_UID_EEPROM_CID, 8);					// Read the SRAM
  uint8_t i=0;
  while(Wire.available()){
#if ARDUINO >= 100
	uid[i] = Wire.read();
#else
	uid[i] = Wire.receive();
#endif
	i++;
  }
}

void MCP794xxRTC::get_eeprom_sequential(uint8_t* buf, uint8_t len){
	if(_part <= 0x02){		// let's simulate an eeprom read (actually gonna read SRAM...)
		read_sram(buf);		// this should be removed from the library
	}						// it's here only because by mistake I order the parts without EEPROM  =(;(-
	if(_part >= 0x10){
		Wire.beginTransmission(MCP794xx_UID_EEPROM_CID);
		#if ARDUINO >= 100  
		Wire.write((uint8_t)_ee_address); 					// Eeprom begins at 0x0
		#else												// don't know how to generate a restart which is an ACK followed by a START
		Wire.send(_ee_address);								// so let's try to set the pointer address first and generate a STOP
		#endif  											// then send control byte to read and ACKs until len transmitted.
		Wire.endTransmission();
	  // request bytes
		if(len > 128){
			Wire.requestFrom(MCP794xx_UID_EEPROM_CID, 128);		// Read the EEPROM, remember the 128 bytes buffer limitation...
		}else{
			Wire.requestFrom(MCP794xx_UID_EEPROM_CID, (int)len);
		}
		uint8_t i=0;										// this need a little work  but don't have enough time...
		while(Wire.available()){
		#if ARDUINO >= 100
			buf[i] = Wire.read();
		#else
			buf[i] = Wire.receive();
		#endif
			i++;
		}
	}
  
} 

// PRIVATE FUNCTIONS

// Convert Decimal to Binary Coded Decimal (BCD)
uint8_t MCP794xxRTC::dec2bcd(uint8_t num)
{
  return ((num/10 * 16) + (num % 10));
}

// Convert Binary Coded Decimal (BCD) to Decimal
uint8_t MCP794xxRTC::bcd2dec(uint8_t num)
{
  return ((num/16 * 10) + (num % 16));
}



